

class TextStream {

    //--------------------------------------------------------

    constructor(contents) {
        this._contents = contents;
        this._pos = 0;
    }

    static from(contents) {
        return new this(contents);
    }

    //--------------------------------------------------------

    sw() {
        //sw = skip white
        while (this.hasNext() && this.atWhite())
            this.skip();
    }

    stw() {
        //stw = skip to white
        while (this.hasNext() && !this.atWhite())
            this.skip();
    }

    sbw() {
        //sbw = skip back over white
        if (this.isEmpty()) return;
        if (this.atEnd()) this.back();
        while (!this.atStart() && this.atWhite())
            this.back();
    }

    sbtw() {
        //sbtw = skip back to white
        while (!this.atStart() && !this.atWhite())
            this.back();
    }

    //--------------------------------------------------------

    //alias'
    ns(match) { return this.nextString(match); }

    nextString(match) {
        /*next string up to match
        if match is undefined, then up to whitespace*/
        if (!this.hasNext()) return '';
        this.sw();
        const pattern = match ? `(\\s+\|${match})` : '\\s+';
        let i2 = this.holdDuring(ss => ss.find(pattern));
        if (i2 < 0) i2 = this.size();
        return this.slice(this.pos(), i2);
    }

    prevString(match) {
        //LIMITATION
        //TODO -- we really need to expand match here (and match backwards)
        //THIS IS CURRENTLY ONLY DEPENDABLE WITH ONE CHAR MATCHES
        /*prev string back to match (inclusive of char at current pot)
        if match is undefined, back to whitespace*/
        //if (!this.hasNext()) return '';
        this.sbw();
        const pattern = match ? `(\\s+\|${match})` : '\\s+';
        let i1 = this.holdDuring(ss => ss.findBack(pattern));
        if (i1 < 0) i1 = 0;
        return this.slice(i1 + 1, this.pos() + 1);
    }

    find(match) {
        //return next matching pos, or -1 (this is mutable method)
        while (this.hasNext()) {
            if (this.nextEquals(match))
                return this.pos();
            this.skip();
        }
        return -1;
    }

    findBack(match) {
        //Return prev matching pos, or -1 (this is mutable method)
        //LIMITATION
        //TODO -- we really need to expand match here (and match backwards)
        //THIS IS CURRENTLY ONLY DEPENDABLE WITH ONE CHAR MATCHES
        while (this.pos() >= 0) {
            if (this.nextEquals(match))
                return this.pos();
            this.back();
        }
        return -1;
    }

    findFor(fct) {
        //return next matching pos, or -1 (this is mutable method)
        //use fct to evaluate each element for match
        while (this.hasNext()) {
            if (fct(this.peek()))
                return this.pos();
            this.skip();
        }
        return -1;
    }

    findForBack(fct) {
        //return prev matching pos, or -1 (this is mutable method)
        //use fct to evaluate each element for match
        while (this.pos() >= 0) {
            if (fct(this.peek()))
                return this.pos();
            this.skip(-1);
        }
        return -1;
    }

    upToFor(fct) {
        //return string from current pos to matching pos (excluding match),
        //if not found return remaining
        //position is after match or end
        //use fct to evaluate each element for match
        const p1 = this.pos();
        this.findFor(fct);
        if (this.hasNext()) {
            this.skip();
            return this.copyFromTo(p1, this.pos() - 2);
        }
        else
            return this.copyFromTo(p1, this.lastPos());
    }

    backToFor(fct) {
        //return string from current pos [back] to matching pos (excluding match),
        //if not found return all previous
        //position is before match or start
        //use fct to evaluate each element for match
        const p2 = this.pos();
        this.findForBack(fct);
        if (this.pos() >= 0) {
            this.skip(-1);
            return this.copyFromTo(this.pos() + 2, p2);
        }
        else
            return this.copyFromTo(0, p2);
    }

    holdDuring(fct) {
        const hold = this.pos();
        try {
            return fct(this);
        } finally {
            this.setPos(hold);
        }
    }

    nextLine() {
        //final pos is beyond '\n'
        if (!this.hasNext())
            return '';
        let line = '';
        const newlinePattern = this.newlinePattern();
        while (this.hasNext() && !this.nextEquals(newlinePattern))
            line += this.next();
        if (this.atNewline())
            this.skip(this.newline().length);
        return line;
    }

    nextEquals(pattern) {
        //assume pattern is in proper format e.g. \\s*
        let chunk;
        const cl = this.constructor;
        //TODO -- add "doHolding"
        const p0 = this.pos();
        //First try chunk (with pattern size)
        chunk = this.copyFromLength(p0, pattern.length * 2);

        if (cl.nextMatchesPattern(chunk, pattern))
            return true;
        //Second try chunk (with triple pattern size)
        chunk = this.copyFromLength(p0, pattern.length * 3);
        if (cl.nextMatchesPattern(chunk, pattern))
            return true;
        //Lastly try chunk (with all)
        chunk = this.copyRemaining();
        return cl.nextMatchesPattern(chunk, pattern);
    }

    repeat(fct, count) {
        for (let i = 0; i < count; i++)
            fct(this);
    }

    //--------------------------------------------------------
    // Moving (generally return true if they succeed)

    moveTo(pattern) {
        while (this.hasNext())
            if (this.nextEquals(pattern))
                return true;
            else
                this.skip();
        return false;
    }

    //--------------------------------------------------------
    //Static Helpers

    static compareWithPattern(string, aPattern) {
        const pattern = `^${aPattern}$`;
        return string.match(pattern) != null;
    }

    static nextMatchesPattern(string, pattern) {
        const match = string.match(this.coerceRegex(pattern));
        return match && match.index === 0;
    }

    static coerceRegex(arg) {
        //arg is pattern or regex
        return isString(arg)
            ? new RegExp(arg)
            : arg;
    }

	static unindent(stringable) {
		let str, lines, counts, indent, unindentedLines;
		str = stringable.toString();
		lines = str.split('\n').filter(line => line.trim().length > 0);
		counts = lines.map(line => {
			const strm = TextStream.from(line);
			strm.findFor(ch => ch.trim().length > 0);
			return strm.pos();
		});
		indent = Math.min(...counts);
		//Now unindent, easy as pi
		unindentedLines = lines.map(line => line.slice(indent));
		return unindentedLines.join('\n');
	}

    //--------------------------------------------------------
    //Copying

    peek() {
        return this.slice(this.pos(), this.pos() + 1);
    }

    slice(p1, p2) {
        //p2 is optional
        //note that js slice is "soft" -- does not throw exception when outside bounds
        return this.contents().slice(p1, p2);
    }

    copyRemaining() {
        return this.slice(this.pos(), this.size());
    }

    copyFromLength(p1, length) {
        return this.slice(p1, p1 + length);
    }

    copyFromTo(p1, p2) {
        return this.slice(p1, p2 + 1);
    }

    //--------------------------------------------------------

    next(count) {
        const n = count ?? 1;
        const str = this.slice(this.pos(), this.pos() + n);
        this.skip(n);
        return str;
    }

    skip(count) {
        const n = count ?? 1;
        this.setPos(this.pos() + n);
    }

    back(count) {
        const n = Math.abs(count ?? 1);
        this.skip(-n);
    }

    //--------------------------------------------------------

    size() {
        return this.contents().length;
    }

    isEmpty() {
        return this.size() === 0;
    }

    atEnd() {
        return !this.hasNext();
    }

    hasNext() {
        return this.pos() <= this.lastPos();
    }

    atStart() {
        return this.pos() === 0;
    }

    lastPos() {
        return this.size() - 1;
    }

    //--------------------------------------------------------

    newline() {
        return '\n';
    }

    newlinePattern() {
        return '\\n';
    }

    atNewline() {
        return this.nextEquals(this.newline());
    }

    atWhite() {
        return this.nextEquals('\\s');
    }

    //--------------------------------------------------------

    reset() {
        this.setPos(0);
    }

    setToEnd() {
        this.setPos(this.size());
    }

    //--------------------------------------------------------

    getContents() { return this._contents; }
    contents() { return this._contents; }
    getPos() { return this._pos; }
    setPos(newPos) { this._pos = newPos; }
    pos(doNotUse) {
        if (doNotUse)
            throw '`use setPos(p) -- not pos(p)';
        return this._pos;
    }

}




TextStream;