'use strict';

class Test {
    /* unitTest
    testLabel
    testFunction*/

constructor(unitTest, testLabel, testFunction) {
        this.unitTest = unitTest;
        this.testLabel = testLabel;
        this.testFunction = testFunction;
		this.count = 0;
		this.failed = 0;
		this.exceptions = 0;
        this.possiblePoints = 100;   //default

        //derived
        this.points = null;
    }

	incrementCount() { this.count++; }
	incrementFailures(){ this.failed++; }
	incrementExceptions(){ this.exceptions++; }

	getTotalCount() {
		//let us assume an exception supressed a count
		return this.count + this.exceptions;
	}

    hasResults() {
        return this.getTotalCount() > 0;
    }

	hasFailureOrExeption() {
		return (this.failed > 0) || (this.exceptions > 0);
	}

	getPassedCount() {
		return Math.max(this.getTotalCount() - this.failed - this.exceptions, 0);
	}

	getPercentPassing() {
		//let us assume an exception supressed a count
		if (this.count === 0) return 0;
		return 100 * this.getPassedCount() / this.count;
	}

    getTestFunction() {
        return this.testFunction;
    }

    getTestLabel() {
        return this.testLabel;
    }

    //---------------------------------------------------------------------------
    // Points = % * possible points

    getPossiblePoints() {
        return this.possiblePoints;
    }

    getPoints() {
        if (this.points == null)
            this.points = this.solvePoints();
        return this.points;
    }

    solvePoints() {
        return Math.round(this.getPossiblePoints() * this.getPercentPassing() / 100);
    }

    getPointsPercent() {
        //could be slightly different than percent because of rounding in solvePoints
        return Math.round(100 * this.getPoints() / this.getPossiblePoints());
    }

    getStatus() {
        return (this.getPoints() >= this.getPossiblePoints()) ? 'passed' : 'FAILED';
    }

    //---------------------------------------------------------------------------

    run() {
        this.unitTest.show("Running -- " + this.testLabel);
        this.unitTest.setCurrentSubTest(this);
        this.testFunction();
        this.unitTest.setCurrentSubTest(null);
    }

    toString() {
        //return `${this.testLabel} -- ${this.getPercentPassing()}% -- ${this.getPassedCount()} of ${this.getTotalCount()}`;
        var points, possible, percent, postfix='';
        points = this.getPoints();
        possible = this.getPossiblePoints();
        percent = this.getPercentPassing();
        if (percent < 100 && (points == possible))
            postfix = ` (actual ${this.getPassedCount()} of ${this.count} or ${percent.toFixed(4)}%)`;
        return `${this.testLabel} -- ${this.getStatus()} -- ${this.getPoints()} of ${this.getPossiblePoints()}` + postfix;
	}
}

//==================================================================

class TestRunner {
    /*unitTest
    subTests*/

    constructor(unitTest, subTests) {
        this.unitTest = unitTest;
        this.subTests = subTests;
    }

	preCheck() {
		const preTestError = this.unitTest.preTest();
		if (!preTestError)
			return true;
        const utest = this.unitTest;
		utest
			.show('PRE-CHECK FAILED')
			.show(`Error: ${preTestError}`)
			.show('Could not run tests')
			.incrementExceptionCount();
		return false;
	}

    runAll() {
		if (!this.preCheck()) return;
        const utest = this.unitTest;
        utest.beforeAll();
        for (let eachSubTest of this.subTests) {

            //utest.beforeEach();
            this.runSubTest(eachSubTest, utest);
            //utest.afterEach();
        }
        utest.afterAll();
        this.unitTest.show("");
		const uname = utest.getUnitName();
		if (!isString(uname))
			name = uname.name;
        utest.printSummary(uname);
    }

    runSubTest(subTest, utest) {
        //xdoc method cascading
		utest.setCurrentSubTest(subTest);
        try {
			utest.beforeEach();
            subTest.run();
            utest.afterEach();
        } catch (ex) {
			const ut = this.unitTest;
            ut
				.incrementExceptionCount()
				.show(`EXCEPTION OCCURRED in ${subTest.toString()}`);
			exceptionToDisplayStrings(ex).forEach(each => ut.show(each));
			ut.show('');
		}
		utest.setCurrentSubTest(null);
	}

	//--------------------------------------------------------

}

//==================================================================

class UnitTestManager {

    /*unitTest;
    subTests*/

    constructor(testee, subTests) {
        this.unitTest = testee;
        this.subTests = subTests;
    }

    getUnitTest() {
        return this.unitTest;
    }

    getSubTests() {
        return this.subTests;
    }

    testRunner() {
        return new TestRunner(this.unitTest, this.subTests);
    }

    runAll() {
        this.testRunner().runAll();
    }

    static runUnitTest(unitTest) {
        let mgr;
        mgr = new UnitTestManager(unitTest, unitTest.getSubTests());
        mgr.runAll();
        return unitTest;
    }

    static runUnitTestWith(unitTest, subTestNames) {
        let mgr, subTests;
		subTests = unitTest.getSubTests();
		subTests = subTests.filter(eaSubTest => subTestNames.includes(eaSubTest.getTestLabel()));
        mgr = new UnitTestManager(unitTest, subTests);
        mgr.runAll();
        return unitTest;
    }

}

//***************************************************
//***************************************************

//----------------------------------------------------
//Helpers

function prn(anObject) {
    let safeValue = anObject;
    if (safeValue === undefined)
        safeValue = '*undefined*';
    if (safeValue === null)
        safeValue = '*null*';
    console.log(safeValue.toString());
}
